#include <marr.h>
#include <string.h>

marr_da_t* _marr_new(size_t stride, size_t cap) {
    marr_da_t* da = malloc(sizeof(marr_da_t));
    da->data = malloc(stride * cap);
    MARR_ASSERT(da->data && "MARR: Could not allocate new dyn array");
    da->cap = cap;
    da->len = 0;
    return da;
}

// AKA end
int _marr_push_back(marr_da_t* da, void* val) {
    MARR_ASSERT(da && "MARR: da is null");
    MARR_ASSERT(val && "MARR: value is null");

    if (marr_len(da) >= marr_cap(da)) {
        if (marr_resize(da, marr_cap(da) * MARR_RESIZE_FACTOR)) {
            return 1;
        }
    }
    memcpy(marr_as_ptr(void*, da) + (marr_len(da) * marr_stride(da)), val, marr_stride(da));
    marr_len(da) += 1;
    return 0;
}

// AKA start
int _marr_push_front(marr_da_t* da, void* val) {
    MARR_ASSERT(da && "MARR: da is null");
    MARR_ASSERT(val && "MARR: value is null");
    if (marr_len(da) >= marr_cap(da)) {
        if (marr_resize(da, marr_cap(da) * MARR_RESIZE_FACTOR)) {
            return 1;
        }
    }
    void* tmp_arr = MARR_MALLOC(marr_len(da) * marr_stride(da));
    MARR_ASSERT(tmp_arr && "MARR: Could not allocate temp_arr buffer");

    memcpy(tmp_arr, marr_as_ptr(void*, da), marr_len(da) * marr_stride(da));
    memcpy(marr_as_ptr(void*, da), val, marr_stride(da));
    memcpy(marr_as_ptr(void*, da) + marr_stride(da), tmp_arr, marr_len(da) * marr_stride(da));
    marr_len(da) += 1;
    return 1;
}

// AKA end
int _marr_pop_back(marr_da_t* da, void* dest) {
    MARR_ASSERT(da && "MARR: da is null");
    if (marr_len(da) < 1) {
        return 1;
    }
    if (dest) {
        memcpy(marr_as_ptr(void*, da) + ((marr_len(da) - 1) * marr_stride(da)), dest, marr_stride(da));
    }

    marr_len(da) -= 1;
    return 0;
}

// AKA start
int _marr_pop_front(marr_da_t* da, void* dest) {
    MARR_ASSERT(da && "MARR: da is null");
    if (dest) {
        memcpy(dest, marr_as_ptr(void*, da), marr_stride(da));
    }
    memcpy(marr_as_ptr(void*, da), marr_as_ptr(void*, da) + marr_stride(da), marr_stride(da) * marr_len(da));
    return 0;
}

int _marr_append(marr_da_t* to, marr_da_t* from) {
    MARR_ASSERT(to && "MARR: da is null");
    MARR_ASSERT(from && "MARR: da is null");
    MARR_ASSERT(marr_stride(to) == marr_stride(from) && "arrays do not have the same type");

    void* tmp = MARR_MALLOC(marr_stride(from));
    for (int i = 0; i < marr_len(from); i++) {
        if (marr_pop_front(from, tmp)) {
            return 1;
        }
        if (marr_push_back(to, tmp)) {
            return 1;
        }
    }

    MARR_FREE(tmp);
    return 1;
}

int _marr_resize(marr_da_t* da, size_t new_cap) {
    MARR_ASSERT(da && "MARR: da is null");
    if (new_cap <= da->cap) {
        return 1;
    }
    da->data = MARR_REALLOC(da->data, new_cap * da->stride);

    MARR_ASSERT(marr_as_ptr(void*, da) && "MARR: Could not resize dyn arr");
    return 0;
}

int _marr_remove(marr_da_t* da, size_t idx) {
    MARR_ASSERT(da && "MARR: da is null");
    void* tmp_arr = MARR_MALLOC((marr_len(da) - 1) * marr_stride(da));
    MARR_ASSERT(tmp_arr && "MARR: Could not allocate temp_arr buffer");
    memcpy(tmp_arr, marr_as_ptr(void*, da), marr_stride(da) * idx);
    memcpy(tmp_arr, marr_as_ptr(void*, da) + (marr_stride(da) * idx + 1), (marr_stride(da) * marr_len(da)) - (marr_stride(da) * idx));
    memcpy(marr_as_ptr(void*, da), tmp_arr, (marr_len(da) - 1) * marr_stride(da));
    MARR_FREE(tmp_arr);
    marr_len(da) -= 1;
    return 0;
}

bool _marr_contains(marr_da_t* da, void* val) {
    MARR_ASSERT(da && "MARR: da is null");
    for (int i = 0; i < marr_len(da); i++) {
        if (memcmp(marr_as_ptr(void*, da) + (marr_stride(da) * i), val, marr_stride(da)) == 0) {
            return true;
        }
    }
    return false;
}

bool _marr_map(marr_da_t* da, marr_da_t* da_out, bool (*func)(void* item, void* data), void* data) {
    MARR_ASSERT(da && "MARR: da is null");
    bool found = false;
    for (int i = 0; i < marr_len(da); i++) {
        if (func(marr_as_ptr(void*, da) + (marr_stride(da) * i), data)) {
            found = true;
            if (da_out) {
                MARR_ASSERT(marr_stride(da) == marr_stride(da_out) && "arrays do not have the same type");
                marr_push_back(da_out, marr_as_ptr(void*, da) + (marr_stride(da) * i));
            }
        }
    }
    return found;
}
